# w.c.s. - web application for online forms
# Copyright (C) 2005-2013  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import json
import sys
import xml.etree.ElementTree as ET

from qommon.form import *
from qommon.misc import http_get_page, http_post_request, get_variadic_url, JSONEncoder
from wcs.workflows import WorkflowStatusItem, register_item_class, template_on_formdata
from wcs.api import sign_url

TIMEOUT = 15

class WebserviceCallStatusItem(WorkflowStatusItem):
    description = N_('Webservice Call')
    key = 'webservice_call'
    support_substitution_variables = True

    url = None
    varname = None
    post = True
    request_signature_key = None
    post_data = None

    def get_parameters(self):
        return ('url', 'post', 'varname', 'request_signature_key', 'post_data')

    def add_parameters_widgets(self, form, parameters, prefix='', formdef=None):
        if 'url' in parameters:
            form.add(StringWidget, '%surl' % prefix,
                     title=_('URL'), value=self.url, size=80,
                     hint=_('Common substitution variables are available with the [variable] syntax.'))
        if 'post' in parameters:
            form.add(CheckboxWidget, '%spost' % prefix,
                    title=_('Post formdata (JSON)'),
                    value=self.post)
        if 'post_data' in parameters:
            form.add(WidgetDict, '%spost_data' % prefix,
                    title=_('Post data'),
                    value=self.post_data or {})
        if 'varname' in parameters:
            form.add(VarnameWidget, '%svarname' % prefix,
                     title=_('Variable Name'), value=self.varname)
        if 'request_signature_key' in parameters:
            form.add(StringWidget, '%srequest_signature_key' % prefix,
                    title=_('Request Signature Key'),
                    value=self.request_signature_key)

    def perform(self, formdata):
        if not self.url:
            # misconfigured action
            return
        url = self.url
        if '[' in url:
            variables = get_publisher().substitutions.get_context_variables()
            url = get_variadic_url(url, variables)

        if self.request_signature_key:
            signature_key = template_on_formdata(formdata, self.request_signature_key)
            if signature_key:
                url = sign_url(url, signature_key)

        headers = {'Content-type': 'application/json',
                'Accept': 'application/json'}
        post_data = None # payload

        # if self.post_data exists, post_data is a dict built from it
        if self.post_data:
            post_data = {}
            for (key, value) in self.post_data.items():
                try:
                    post_data[key] = self.compute(value, raises=True)
                except:
                    get_publisher().notify_of_exception(sys.exc_info())

        # if formdata has to be sent, it's the payload. If post_data exists,
        # it's added in formdata['extra']
        if self.post:
            formdata_dict = formdata.get_json_export_dict()
            if post_data is not None:
                formdata_dict['extra'] = post_data
            post_data = formdata_dict

        if post_data is not None:
            post_data = json.dumps(post_data, cls=JSONEncoder,
                    encoding=get_publisher().site_charset)
            response, status, data, authheader = http_post_request(
                    url, post_data, headers=headers, timeout=TIMEOUT)
        else:
            response, status, data, auth_header = http_get_page(
                    url, headers=headers, timeout=TIMEOUT)

        if self.varname:
            workflow_data = {'%s_status' % self.varname: status}
            if status == 200:
                try:
                    d = json.loads(data)
                except (ValueError, TypeError):
                    pass
                else:
                    workflow_data['%s_response' % self.varname] = d
                    if isinstance(d.get('data'), dict) and d['data'].get('display_id'):
                        formdata.id_display = d.get('data', {}).get('display_id')
                    elif d.get('display_id'):
                        formdata.id_display = d.get('display_id')
            formdata.update_workflow_data(workflow_data)
            formdata.store()

        if (status // 100) not in (2, 3):
            raise Exception("Call to webservice gave %s as status code" % status)

    def post_data_export_to_xml(self, xml_item, charset, include_id=False):
        if not self.post_data:
            return
        el = ET.SubElement(xml_item, 'post_data')
        for (key, value) in self.post_data.items():
            item = ET.SubElement(el, 'item')
            if type(key) is unicode:
                ET.SubElement(item, 'name').text = key
            elif type(key) is str:
                ET.SubElement(item, 'name').text = unicode(key, charset, 'replace')
            else:
                raise AssertionError('unknown type for key (%r)' % key)
            if type(value) is unicode:
                ET.SubElement(item, 'value').text = value
            elif type(value) is str:
                ET.SubElement(item, 'value').text = unicode(value, charset, 'replace')
            else:
                raise AssertionError('unknown type for value (%r)' % key)


    def post_data_init_with_xml(self, elem, charset, include_id=False):
        if elem is None:
            return
        self.post_data = {}
        for item in elem.findall('item'):
            key = item.find('name').text.encode(charset)
            value = item.find('value').text.encode(charset)
            self.post_data[key] = value

register_item_class(WebserviceCallStatusItem)
